package org.eclipse.jdt.internal.compiler.classfmt;

package .tmp;

import java.util.Arrays;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileStruct;
import org.eclipse.jdt.internal.compiler.classfmt.ElementValuePairInfo;
import org.eclipse.jdt.internal.compiler.codegen.ConstantPool;
import org.eclipse.jdt.internal.compiler.env.ClassSignature;
import org.eclipse.jdt.internal.compiler.env.EnumConstantSignature;
import org.eclipse.jdt.internal.compiler.env.IBinaryAnnotation;
import org.eclipse.jdt.internal.compiler.env.IBinaryElementValuePair;
import org.eclipse.jdt.internal.compiler.impl.BooleanConstant;
import org.eclipse.jdt.internal.compiler.impl.ByteConstant;
import org.eclipse.jdt.internal.compiler.impl.CharConstant;
import org.eclipse.jdt.internal.compiler.impl.DoubleConstant;
import org.eclipse.jdt.internal.compiler.impl.FloatConstant;
import org.eclipse.jdt.internal.compiler.impl.IntConstant;
import org.eclipse.jdt.internal.compiler.impl.LongConstant;
import org.eclipse.jdt.internal.compiler.impl.ShortConstant;
import org.eclipse.jdt.internal.compiler.impl.StringConstant;
import org.eclipse.jdt.internal.compiler.util.Util;

public class AnnotationInfo extends ClassFileStruct implements IBinaryAnnotation {
  private char[] typename;
  
  private ElementValuePairInfo[] pairs;
  
  long standardAnnotationTagBits = 0L;
  
  int readOffset = 0;
  
  static Object[] EmptyValueArray = new Object[0];
  
  AnnotationInfo(byte[] classFileBytes, int[] contantPoolOffsets, int offset) {
    super(classFileBytes, contantPoolOffsets, offset);
  }
  
  AnnotationInfo(byte[] classFileBytes, int[] contantPoolOffsets, int offset, boolean runtimeVisible, boolean populate) {
    this(classFileBytes, contantPoolOffsets, offset);
    if (populate) {
      decodeAnnotation();
    } else {
      this.readOffset = scanAnnotation(0, runtimeVisible, true);
    } 
  }
  
  private void decodeAnnotation() {
    this.readOffset = 0;
    int utf8Offset = this.constantPoolOffsets[u2At(0)] - this.structOffset;
    this.typename = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
    int numberOfPairs = u2At(2);
    this.readOffset += 4;
    this.pairs = (numberOfPairs == 0) ? ElementValuePairInfo.NoMembers : new ElementValuePairInfo[numberOfPairs];
    for (int i = 0; i < numberOfPairs; i++) {
      utf8Offset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
      char[] membername = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
      this.readOffset += 2;
      Object value = decodeDefaultValue();
      this.pairs[i] = new ElementValuePairInfo(membername, value);
    } 
  }
  
  Object decodeDefaultValue() {
    char[] typeName, constName, className;
    int numberOfValues;
    Object value = null;
    int tag = u1At(this.readOffset);
    this.readOffset++;
    int constValueOffset = -1;
    switch (tag) {
      case 90:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = BooleanConstant.fromValue((i4At(constValueOffset + 1) == 1));
        this.readOffset += 2;
        return value;
      case 73:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = IntConstant.fromValue(i4At(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 67:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = CharConstant.fromValue((char)i4At(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 66:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = ByteConstant.fromValue((byte)i4At(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 83:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = ShortConstant.fromValue((short)i4At(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 68:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = DoubleConstant.fromValue(doubleAt(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 70:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = FloatConstant.fromValue(floatAt(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 74:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = LongConstant.fromValue(i8At(constValueOffset + 1));
        this.readOffset += 2;
        return value;
      case 115:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        value = StringConstant.fromValue(String.valueOf(utf8At(constValueOffset + 3, u2At(constValueOffset + 1))));
        this.readOffset += 2;
        return value;
      case 101:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        typeName = utf8At(constValueOffset + 3, u2At(constValueOffset + 1));
        this.readOffset += 2;
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        constName = utf8At(constValueOffset + 3, u2At(constValueOffset + 1));
        this.readOffset += 2;
        value = new EnumConstantSignature(typeName, constName);
        return value;
      case 99:
        constValueOffset = this.constantPoolOffsets[u2At(this.readOffset)] - this.structOffset;
        className = utf8At(constValueOffset + 3, u2At(constValueOffset + 1));
        value = new ClassSignature(className);
        this.readOffset += 2;
        return value;
      case 64:
        value = new org.eclipse.jdt.internal.compiler.classfmt.AnnotationInfo(this.reference, this.constantPoolOffsets, this.readOffset + this.structOffset, false, true);
        this.readOffset += ((org.eclipse.jdt.internal.compiler.classfmt.AnnotationInfo)value).readOffset;
        return value;
      case 91:
        numberOfValues = u2At(this.readOffset);
        this.readOffset += 2;
        if (numberOfValues == 0) {
          value = EmptyValueArray;
        } else {
          Object[] arrayElements = new Object[numberOfValues];
          value = arrayElements;
          for (int i = 0; i < numberOfValues; i++)
            arrayElements[i] = decodeDefaultValue(); 
        } 
        return value;
    } 
    throw new IllegalStateException("Unrecognized tag " + (char)tag);
  }
  
  public IBinaryElementValuePair[] getElementValuePairs() {
    if (this.pairs == null)
      initialize(); 
    return (IBinaryElementValuePair[])this.pairs;
  }
  
  public char[] getTypeName() {
    return this.typename;
  }
  
  void initialize() {
    if (this.pairs == null)
      decodeAnnotation(); 
  }
  
  private int readRetentionPolicy(int offset) {
    int utf8Offset;
    char[] typeName;
    int numberOfValues, i, currentOffset = offset;
    int tag = u1At(currentOffset);
    currentOffset++;
    switch (tag) {
      case 101:
        utf8Offset = this.constantPoolOffsets[u2At(currentOffset)] - this.structOffset;
        typeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
        currentOffset += 2;
        if (typeName.length == 38 && CharOperation.equals(typeName, ConstantPool.JAVA_LANG_ANNOTATION_RETENTIONPOLICY)) {
          utf8Offset = this.constantPoolOffsets[u2At(currentOffset)] - this.structOffset;
          char[] constName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
          this.standardAnnotationTagBits |= Annotation.getRetentionPolicy(constName);
        } 
        currentOffset += 2;
        return currentOffset;
      case 66:
      case 67:
      case 68:
      case 70:
      case 73:
      case 74:
      case 83:
      case 90:
      case 99:
      case 115:
        currentOffset += 2;
        return currentOffset;
      case 64:
        currentOffset = scanAnnotation(currentOffset, false, false);
        return currentOffset;
      case 91:
        numberOfValues = u2At(currentOffset);
        currentOffset += 2;
        for (i = 0; i < numberOfValues; i++)
          currentOffset = scanElementValue(currentOffset); 
        return currentOffset;
    } 
    throw new IllegalStateException();
  }
  
  private int readTargetValue(int offset) {
    int utf8Offset;
    char[] typeName;
    int numberOfValues, currentOffset = offset;
    int tag = u1At(currentOffset);
    currentOffset++;
    switch (tag) {
      case 101:
        utf8Offset = this.constantPoolOffsets[u2At(currentOffset)] - this.structOffset;
        typeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
        currentOffset += 2;
        if (typeName.length == 34 && CharOperation.equals(typeName, ConstantPool.JAVA_LANG_ANNOTATION_ELEMENTTYPE)) {
          utf8Offset = this.constantPoolOffsets[u2At(currentOffset)] - this.structOffset;
          char[] constName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
          this.standardAnnotationTagBits |= Annotation.getTargetElementType(constName);
        } 
        currentOffset += 2;
        return currentOffset;
      case 66:
      case 67:
      case 68:
      case 70:
      case 73:
      case 74:
      case 83:
      case 90:
      case 99:
      case 115:
        currentOffset += 2;
        return currentOffset;
      case 64:
        currentOffset = scanAnnotation(currentOffset, false, false);
        return currentOffset;
      case 91:
        numberOfValues = u2At(currentOffset);
        currentOffset += 2;
        if (numberOfValues == 0) {
          this.standardAnnotationTagBits |= 0x800000000L;
        } else {
          for (int i = 0; i < numberOfValues; i++)
            currentOffset = readTargetValue(currentOffset); 
        } 
        return currentOffset;
    } 
    throw new IllegalStateException();
  }
  
  private int scanAnnotation(int offset, boolean expectRuntimeVisibleAnno, boolean toplevel) {
    int currentOffset = offset;
    int utf8Offset = this.constantPoolOffsets[u2At(offset)] - this.structOffset;
    char[] typeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
    if (toplevel)
      this.typename = typeName; 
    int numberOfPairs = u2At(offset + 2);
    currentOffset += 4;
    if (expectRuntimeVisibleAnno && toplevel)
      switch (typeName.length) {
        case 22:
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_DEPRECATED)) {
            this.standardAnnotationTagBits |= 0x400000000000L;
            return currentOffset;
          } 
          break;
        case 23:
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_SAFEVARARGS)) {
            this.standardAnnotationTagBits |= 0x8000000000000L;
            return currentOffset;
          } 
          break;
        case 29:
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_ANNOTATION_TARGET)) {
            currentOffset += 2;
            return readTargetValue(currentOffset);
          } 
          break;
        case 32:
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_ANNOTATION_RETENTION)) {
            currentOffset += 2;
            return readRetentionPolicy(currentOffset);
          } 
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_ANNOTATION_INHERITED)) {
            this.standardAnnotationTagBits |= 0x1000000000000L;
            return currentOffset;
          } 
          break;
        case 33:
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_ANNOTATION_DOCUMENTED)) {
            this.standardAnnotationTagBits |= 0x800000000000L;
            return currentOffset;
          } 
          break;
        case 52:
          if (CharOperation.equals(typeName, ConstantPool.JAVA_LANG_INVOKE_METHODHANDLE_POLYMORPHICSIGNATURE)) {
            this.standardAnnotationTagBits |= 0x10000000000000L;
            return currentOffset;
          } 
          break;
      }  
    for (int i = 0; i < numberOfPairs; i++) {
      currentOffset += 2;
      currentOffset = scanElementValue(currentOffset);
    } 
    return currentOffset;
  }
  
  private int scanElementValue(int offset) {
    int numberOfValues, i, currentOffset = offset;
    int tag = u1At(currentOffset);
    currentOffset++;
    switch (tag) {
      case 66:
      case 67:
      case 68:
      case 70:
      case 73:
      case 74:
      case 83:
      case 90:
      case 99:
      case 115:
        currentOffset += 2;
        return currentOffset;
      case 101:
        currentOffset += 4;
        return currentOffset;
      case 64:
        currentOffset = scanAnnotation(currentOffset, false, false);
        return currentOffset;
      case 91:
        numberOfValues = u2At(currentOffset);
        currentOffset += 2;
        for (i = 0; i < numberOfValues; i++)
          currentOffset = scanElementValue(currentOffset); 
        return currentOffset;
    } 
    throw new IllegalStateException();
  }
  
  public String toString() {
    StringBuffer buffer = new StringBuffer();
    buffer.append('@');
    buffer.append(this.typename);
    if (this.pairs != null) {
      buffer.append('(');
      buffer.append("\n\t");
      for (int i = 0, len = this.pairs.length; i < len; i++) {
        if (i > 0)
          buffer.append(",\n\t"); 
        buffer.append(this.pairs[i]);
      } 
      buffer.append(')');
    } 
    return buffer.toString();
  }
  
  public int hashCode() {
    int result = 1;
    result = 31 * result + Util.hashCode((Object[])this.pairs);
    result = 31 * result + CharOperation.hashCode(this.typename);
    return result;
  }
  
  public boolean equals(Object obj) {
    if (this == obj)
      return true; 
    if (obj == null)
      return false; 
    if (getClass() != obj.getClass())
      return false; 
    org.eclipse.jdt.internal.compiler.classfmt.AnnotationInfo other = (org.eclipse.jdt.internal.compiler.classfmt.AnnotationInfo)obj;
    if (!Arrays.equals((Object[])this.pairs, (Object[])other.pairs))
      return false; 
    if (!Arrays.equals(this.typename, other.typename))
      return false; 
    return true;
  }
}
